use strict ;

#Display differences in file lists and file size between remote and local arborescence using ftp. 
# Use option -h for complete documentation.
#Copyright (c) Baptiste Marcel (baptiste@dunwich.org) under GNU GPL, refer to complete documentation documentation for more details.

use File::Basename;
use File::Copy;
use Crypt::RC4;  #require specific package 

#[-L <connexion> -p <cnx pwd> -l <cnx login>]
my $STD_ScriptName = basename($0) ;
my $HelpUsage="Usage: $STD_ScriptName [-{d|D}s] [-u[H]] {-S <profile>|-f <config file>} [-P <passwd>] [-x <key>] [-e dir1[,dir2]]
       $STD_ScriptName -h
       $STD_ScriptName -v
       $STD_ScriptName -P <passwd> -X <key>" ;

my $STD_Version='2.20' ;
my $CopyrightNotice="Copyright (c) Dr Baptiste Marcel 2002 and later.\nThis programms is freeware, you can modify and/or distribute them under the term of the GNU General Public License version 2 or later. It comes with no warranty of any sort. Refer to www.gnu.org/copyleft/gpl.html for the full text of the GNU GPL.";
my $fModeSilent;
       
sub print_msg
  {
  my ($msg, $level, $OutHandler) = ('', 0, *STDOUT);
  if ($#_ + 1 ge 1)
    {$msg=$_[0];}
  if ($#_ + 1 ge 2)
    {$level=$_[1];}
  if ($#_ + 1 ge 3)
    {$OutHandler=$_[2];}

  if ( ! $fModeSilent || $level gt 0 )
    { print $OutHandler $msg; }
  }

sub bHelp
  {
  print "$HelpUsage\n" ;
  print "Display differences in file lists and file size between remote and local arborescence using ftp. FTP connection information can be set as known profiles (included in the guts of the script, crypted or absent password recommended) or in configuration files.\n";
  print "Available profiles in the guts: freebiert and freedunwich
  
  A local file which is not in the remote directory is refered as a sterile file, and is uploaded with option -u.
  A remote file which is not in the local directory is refered as a orphean file, and is deleted with options -u and -d

  The excluded directories, either specified by -e or in the config file, will be ignored (treated as if already synchronised), but will be created if missing (with option -u).
  Excluded list should be a list of coma separated subdir, path relative from remote_root or MainDir, without leading './', with '/' ad subdir separator if required.
  
  
Options:
  -f: use connection information from config file. Syntax for the config file is (password is optional):
remote_host	=ftp server
remote_user	=ftp user
remote_pwd 	=password  
remote_root	=directory to synchronise (use absolute path)
MainDir		=directory used as reference on local host (use absolute path)
exclude         =exclude list (see above)

 -x: with this option, the script assumes that password is crypted with <key>. This options is useful if your config files or script source get distributed, for example if the config file is in the synchronized directory.

 -u: update. Sterile files will be uploaded. This option is not set as default for safety purposes.
 -d: delete. Combine with option -u: orphean files will be deleted. This option is not set as default with -u for safety purposes. If the orphean is a directory not empty, it will probably not be deleted unless you use option -D. Remote deltree are not suported by default, mainly bacause of the danger they represent. Most likely, orphean not-empty directories are the result of tree reorganisation and a remote rename is far a better action to perform.
 -D: include deletion of not-empty directory trees. Only use this option as last chance. If -D is set, option -d is implicit.
 -H: convert local *.htm{l?} files dos to unix before sending them in binary mode
 -e: exclude list (see above)

 -p: specify proxy

 -X: with this option, encrypt the password given by option -P with <key> and display both. It does nothing else. Use it to set your config files with encrypted passwords.
 -h: display this help
 -v: display version
 -s: mode silent: suppress info, display only warnings and errors

Note:
  As this program is designed and developped on DOS-based platform, local file paths are processed as lower case, and you should only use lower case for remote platform file paths. Results with case-sensitive local platform are unpredictable, and there is no planned upgrade for this restriction as long as I am not able to develop on case sensitive platform. If you need case-sensitive suport, consider this as a chance for you to contribute to the software upgrade.

 $STD_ScriptName is copyright (c) Baptiste Marcel 2001-2002.
 $STD_ScriptName is a freeware, you can modify and/or distribute it under the term of the GNU General Public License version 2 or later. It comes with no warranty of any sort. Refer to www.gnu.org/copyleft/gpl.html for the full text of the GNU GPL.
 Always make sure you use the last version which is available at http://www.dunwich.org/soft
 Bug reports and feature requests should be sent to author at baptiste\@dunwich.org.
 Refer to http://www.dunwich.org for more contact information about the author.
\n" ;
  print "$STD_ScriptName version $STD_Version\n";
  }

use Net::FTP;
use Getopt::Std;

our ($opt_D, $opt_s, $opt_d, $opt_u, $opt_h, $opt_H, $opt_S, $opt_f, $opt_P, $opt_x, $opt_X, $opt_v, $opt_e) = 
     (0,      0,       0,       0,      0,       "",     "",   "",       '',    '',    0,      '',   ''  ) ;
getopts('DsudhS:Hf:P:x:X:ve:') || die ("Erreur dans les paramtres d'entre\n$HelpUsage");
if ($opt_h)
  {
  bHelp ;
  exit ;
  }

if ($opt_v)
  {
  print "$STD_Version\n" ;
  exit ;
  }

# my $DialupProg = "$ENV{progra}\\dialup.exe";
##my ($dialup_cnx_name, $dialup_cnx_user, $dialup_cnx_pwd);
if ($opt_X && ! $opt_P) { die "Option -X require option -P\n$HelpUsage";}
#if ($opt_L && (! $opt_l || ! $opt_p)) { die "Option -L require option -p and -l\n$HelpUsage";}
#if ($opt_L && $opt_l && $opt_p)
#  {
#  $dialup_cnx_name=$opt_L;
#  $dialup_cnx_user=$opt_L;
#  $dialup_cnx_pwd=$opt_L;
#  ( -x $DialupProg ) || die "Option -L require tool $DialupProg.";
#  $dialup_cnx_name =~ s/^"(.+)"$/$1 ;
#  $dialup_cnx_user =~ s/^"(.+)"$/$1 ;
#  $dialup_cnx_pwd =~ s/^"(.+)"$/$1 ;
#  }
if ($opt_X && $opt_P)
  { print "$opt_P -> ", b_encrypt( $opt_X, $opt_P ), "\n"; exit ;  }
  
my ($fModeCrypt, $CryptKey, $excluded_list) = (0, '', '');
if ( $opt_e )
  { $excluded_list=$opt_e ; }

if ( $opt_x )
  { $fModeCrypt = 1 ; $CryptKey=$opt_x ; }

$fModeSilent = $opt_s ;

my $fModeUpdate = $opt_u ;
my $fModeDelete = $opt_d || $opt_D ;
my $fModeDeleteDeltree = $opt_D ;
my $fChangeDos2Unix = $opt_H ;

( $opt_H && ! $opt_u ) && die "Option -H requires option -u\n$HelpUsage";

my $profilFTP='' ;
if ( $opt_S ne '' )
  {
  $profilFTP=$opt_S ;
  }
my($ConfigFile, $Password) ;
  
if ( $opt_f ne '' )
  {
  $ConfigFile=$opt_f ;
  }
if ( $opt_P ne '' )
  {
  $Password=$opt_P ;
  }
else { $Password='' ; }  
( $opt_f eq '' || $opt_S eq '' ) || die "Options -S and -f exclusive.";


my (%profil_freedunwich, %profil_freebiert, %profil_freedun_washdc) ;

$profil_freedun_washdc{remote_host}="ftpperso.free.fr";
$profil_freedun_washdc{remote_user}="dunwich";
if ( $fModeCrypt ) { $profil_freedun_washdc{remote_pwd}=b_decrypt($CryptKey,'foo_bar'); }
else               { $profil_freedun_washdc{remote_pwd}=''; }
$profil_freedun_washdc{remote_root}="/washdc";
$profil_freedun_washdc{local_root}= 'c:\data\media\photos\washington' ;

$profil_freedunwich{remote_host}="ftpperso.free.fr";
$profil_freedunwich{remote_user}="dunwich";
if ( $fModeCrypt ) { $profil_freedunwich{remote_pwd}=b_decrypt($CryptKey,'bar'); }
else               { $profil_freedunwich{remote_pwd}=''; }
$profil_freedunwich{remote_root}="/thai";
$profil_freedunwich{local_root}= 'd:\bma\soc\country\th\album' ;

$profil_freebiert{remote_host}="ftpperso.free.fr";
$profil_freebiert{remote_user}="biert";
if ( $fModeCrypt ) { $profil_freebiert{remote_pwd}=b_decrypt($CryptKey,'foobar'); }
else               { $profil_freebiert{remote_pwd}='coco'; }
$profil_freebiert{remote_root}="/album_2001-07-17";
$profil_freebiert{local_root}= 'd:\bma\soc\country\biert\2001-07-17' ;


my %Excluded ;

my $e=$excluded_list;
while (s/^([^,]+),//)
  {
  $Excluded{"./$1"}=1;
  }
$Excluded{"./$e"}=1 if ( $e ne '' );


#print("List of exclusion: |", keys(%Excluded), "|\n");


my %profil;
my $remote_host  ; 
my $remote_user  ; 
my $remote_pwd  ; 
my $remote_root  ; 
my $local_root  ; 
if ( $profilFTP ne '' )
  {
  if ( $profilFTP eq 'freebiert' )
    { %profil = %profil_freebiert ; }
  elsif ( $profilFTP eq 'freedunwich' )
    { %profil = %profil_freedunwich ; }
  elsif ( $profilFTP eq 'freedun_washdc' )
    { %profil = %profil_freedun_washdc ; }
  else
    { die "Unknown profile.\n" ; }
  $remote_host = $profil{remote_host} ; 
  $remote_user = $profil{remote_user} ; 
  $remote_pwd  = $profil{remote_pwd} ; 
  $remote_root = $profil{remote_root} ; 
  $local_root  = $profil{local_root} ; 

  }
else
  {
  ( -f  $ConfigFile ) || die "File $ConfigFile does not exist." ;
  #......................................................................
  #....       Rcupration des paramtres du fichier de config       ....
  #......................................................................
  my (%ConfigTable) ;
  open (CONFIG_FILE, "< $ConfigFile") || die "Unable to open $ConfigFile";
  while ( <CONFIG_FILE> )
    {
    chomp ;
    my($VarName, $VarValue) ;
  
    m/^([^	 =]+)[ 	]*=(.*)$/ ;
    $VarName = $1 ;
    $VarValue = $2 ;
    
#Shell variable interpolation
#If value contains variable (such as ${TEMP}), then replace ${ENV} by its value

  $VarValue =~ s/(\$\{([^}]+)\})/$ENV{$2}/g ;

    $ConfigTable{$VarName} = $VarValue ;
  #  print ("$VarName: $VarValue ;\n");
    }
  close (CONFIG_FILE) ;
  $remote_host = $ConfigTable{remote_host} ; 
  $remote_user = $ConfigTable{remote_user} ; 
  if ( $fModeCrypt ) { $remote_pwd=b_decrypt($CryptKey,$ConfigTable{remote_pwd}); }
  else               { $remote_pwd=$ConfigTable{remote_pwd}; }
  $remote_root = $ConfigTable{remote_root} ; 
  $local_root  = $ConfigTable{MainDir} ; 
  ( $remote_host ne '' && $remote_user ne '' && $remote_root ne '' && $local_root ne '' ) || 
    die "Some missing parameters in $ConfigFile. I need: remote_host, remote_user, remote_root, MainDir." ;
  ( $remote_root =~ m/^\// ) || die "Remote dir '$remote_root' must be an absolute directory, otherwise, the script goes nuts. I will save you tyhis trouble and die now." ;

  $e=$ConfigTable{exclude};
  while (s/^([^,]+),//)
    {
    $Excluded{"./$1"}=1;
    }
  $Excluded{"./$e"}=1 if ( $e ne '' );
  }
  


#Compare deux rpertoires: un sur FTP (/tai/ref sur dunwich.fr) et un autre local

if ( $Password ne '' ) { $remote_pwd=$Password ; }

my $ftp;


$ftp = Net::FTP->new($remote_host, Debug => 0) || die "Unable to initiate $remote_host.";
$ftp->login($remote_user,$remote_pwd) || die "Unable to connect $remote_user\@ftp://$remote_host with pass=$remote_pwd $@" ;
$ftp->binary ;

my $fUpdated ;

do 
  {
  $fUpdated = 0 ;
  print_msg ("Starting $remote_user\@ftp://$remote_host$remote_root/. with pass=<hiden>\n");  # $remote_pwd
  CompareRecurseRemoteLocal ("."); }
until ( (! $fUpdated)  || (! $fModeUpdate) ) ; #Boucle tant qu'il y a des updates
  
  
$ftp->quit;


sub CompareRecurseRemoteLocal
  {
  my $rl_path=$_[0] ;

#  print_msg ("suggesting is excluded $rl_path  \n");
  return   print_msg ("    compared $rl_path: excluded\n") if ($Excluded{"$rl_path"});
  
  my $remote_dir = lc("$remote_root/$rl_path") ;
  my $local_dir = lc("$local_root\\$rl_path") ;
  
  if ( ! $ftp->cwd($remote_dir) )
    {
    print "Creating remote root $remote_dir.\n";
    $ftp->mkdir($remote_dir) || die "Unable to create $remote_dir: $@\n";
    $ftp->cwd($remote_dir)
    }
    
  opendir (LOCAL_ROOT, $local_dir) || die "Rpertoire $local_dir inaccessible" ;
  my @LocalRawFilelist = readdir(LOCAL_ROOT) ;
  close (LOCAL_ROOT) ;
    
  my @LocalFileList = () ; my %LocalLongFileList ;

  foreach my $d ( @LocalRawFilelist )
    {
    if ( $d ne '.' && $d ne '..' )
      { 
      $d = lc($d) ;
      push @LocalFileList, $d ;
      my $F = {} ;
      $F->{filename} = $d ;
      $LocalLongFileList{$d} = $F ;
      }
    }  
  @LocalFileList = sort @LocalFileList ;


#Le dir longlist est une liste de hachoirs
#Doc:
#  RemoteRawFilelist: liste remote en mode raw
#  RemoteLongFileList: hachoir index sur le filename minuscule qui contient une ref de structure avec toutes les proprits.

  #Out of this list, we need two resulting structures:
  #  @RemoteFileList : list of file names (leafname)
  #  $to_RemoteLongFileList : ref top hash indexed by leafname that contains hash that details the properties of the file



  my ($to_RemoteFileList, $to_RemoteLongFileList)=read_ftp_dir($ftp, '.');

  my @RemoteFileList = @$to_RemoteFileList ; 
  my %RemoteLongFileList = %{$to_RemoteLongFileList};
  
  my $NbDeDiff = 0;
  my $DiffMsg;
  if ( $#RemoteFileList != $#LocalFileList )
    { 
    $DiffMsg =  "$#RemoteFileList != $#LocalFileList" ; 
    my @SynchroFileList = () ;
    foreach my $F (@LocalFileList)
      {
      ## $F is the leafname of file
      
      if ( ! defined $RemoteLongFileList{$F} )
        { 
	print "Sterile $rl_path/$F\n";
	if ( $fModeUpdate )
	  {
	  $ftp->cwd($remote_dir);
	  chdir($local_dir);
	  if ( -f $F )
	    {
	    print "  sending $rl_path/$F\n" ;
	    $ftp->put($F,lc($F)) || print STDERR "Unable to send $F\n";
	    $fUpdated++ ;
	    }
	  elsif ( -d $F )
	    {
	    print "creating $F on $remote_dir\n" ;
	    ( $ftp->mkdir(lc($F)) &&
	      CompareRecurseRemoteLocal ("$rl_path/$F") ) || 
	        print STDERR "Unable to create $F\n";
	    }
	  }
	}
      else
        { push @SynchroFileList, $F }
      }
    foreach my $F (@RemoteFileList)
      {
      if ( ! defined $LocalLongFileList{$F} )
        { 
	print "Orphean path $rl_path/$F\n"; 
	if ( $fModeDelete )
	  {
          my $rights = $RemoteLongFileList{$F}->{rights} ;
          if ( $rights =~ /^d/ && $fModeDeleteDeltree )
            { 
            print "  removing tree $rl_path/$F\n" ;
            ftp_deltree($ftp, "$remote_dir/$F"); 
            }
          else
            {
	    $ftp->cwd($remote_dir);
            print "  removing path $rl_path/$F\n" ;
            $ftp->delete($F) || print STDERR "Unable to delete file $F.\n";
            }
	  $fUpdated++ ;
          }
	}

      }      
    }
  else
    {
    foreach my $f (0..$#RemoteFileList)
      {
      my $RE=$LocalFileList[$f] ;
      if ( ! ($RemoteFileList[$f] =~ /^$RE$/i) )
        {
	$NbDeDiff++ ; 
	#Le fichier remote est diffrent du fichier local.
	#Ca ne veut pas forcment dire que c'est un strile face  un orphelin. C'est peut-tre un dcalage.
	#Il faut savoir si le remote est orphelin.
	if ( defined $LocalLongFileList{$RemoteFileList[$f]} )
	  { #Le remote n'est pas orphelin, juste mal plac. 
          print "Misplaced remote $RemoteFileList[$f]\n"; }
	else
	  { #Le remote n'existe pas, il est orphelin.

	  print "Orphean path $rl_path/$RemoteFileList[$f]\n"; 

          if ( $fModeDelete )
            {
            my $F=$RemoteFileList[$f];
            my $rights = $RemoteLongFileList{$F}->{rights} ;
            if ( $rights =~ /^d/ && $fModeDeleteDeltree )
              { 
              print "  removing tree $rl_path/$F\n" ;
              ftp_deltree($ftp, "$remote_dir/$F"); 
              }
            else
              {
              $ftp->cwd($remote_dir);
              print "  removing path $rl_path/$F\n" ;
              $ftp->delete($F) || print STDERR "Unable to delete file $F.\n";
              }
            $fUpdated++ ;
            }
	  }

	#Ca peut aussi tre un strile.
	if ( defined $RemoteLongFileList{$LocalFileList[$f]} )
	  {
	  #Misplaced
          print "Misplaced local $LocalFileList[$f]\n"; }
	else
	  {
	  my $F=$LocalFileList[$f];
	  #Sterile
    	  print "Sterile $rl_path/$F\n";
    	  if ( $fModeUpdate )
    	    {
    	    $ftp->cwd($remote_dir);
    	    chdir($local_dir);
    	    if ( -f $F )
    	      {
    	      print "  sending $rl_path/$F\n" ;
    	      $ftp->put($F,lc($F)) || print STDERR "Unable to send $F";
    	      $fUpdated++ ;
    	      }
    	    elsif ( -d $F )
    	      {
    	      print "creating $F on $remote_dir\n" ;
    	      ( $ftp->mkdir(lc($F)) &&
    	        CompareRecurseRemoteLocal ("$rl_path/$F") ) || 
    	          print STDERR "Unable to create $F";
    	      }
    	    }
    	  }


	}
      else
        { 
	my $rights = $RemoteLongFileList{$RemoteFileList[$f]}->{rights} ;
	if ( $rights =~ /^d/ )
	  { CompareRecurseRemoteLocal ("$rl_path/$RemoteFileList[$f]");  }
	else
	  {
	  my $RemoteSize = $RemoteLongFileList{$RemoteFileList[$f]}->{size};
          my @FileStats = stat("$local_dir\\$RemoteFileList[$f]");
	  my $LocalSize = $FileStats[7];
	  if ( $RemoteSize != $LocalSize )
	    { 
	    print "$rl_path/$RemoteFileList[$f]\tsize $LocalSize != $RemoteSize.\n" ; 
	    if ( $fModeUpdate )
	      {
#En mode update, avec l'option -H: si c'est un fichier HTML, on le convertit en mode unix avant de l'envoyer en binaire
	      $ftp->cwd($remote_dir);
	      chdir($local_dir);
              if ( $RemoteFileList[$f] =~ /\.html?$/ && $fChangeDos2Unix )
	        { 
	        print "  removing 0x0D," ;
		dos2unix ($RemoteFileList[$f]) ;
		}
	      print "  sending\n" ;
	      $ftp->put($RemoteFileList[$f]) || print STDERR "Unable to send $RemoteFileList[$f]";;
	      $fUpdated++ ;
	      }
            }
	  }
        }
      }
    $DiffMsg = "diff=$NbDeDiff/$#RemoteFileList" ;
  
    }

  print_msg ("    compared $rl_path: $DiffMsg\n");

  }  

sub dos2unix
  {
  my $SrcPath = $_[0];
  my $TgtPath= "$SrcPath.out" ;
  open (IN, "< $SrcPath") || die "Unable to read from $SrcPath.\n" ;
  open (OUT, "> $TgtPath") || die "Unable to write in $TgtPath.\n" ;
  binmode IN, ":crlf" ;
  binmode OUT, ":raw" ;
  while (<IN>)
    { 
      $_ =~ s/\x0D//g ;
      print (OUT "$_");
    }
  close (IN);  
  close (OUT);  

  ( -f $SrcPath && -f $TgtPath ) || die "Missing $SrcPath or $TgtPath file.\n" ;
  unlink ($SrcPath) && move ($TgtPath, $SrcPath) ;
  }

sub b_encrypt
  { 
  my $digitstring='';
  my $word8b=RC4($_[0],$_[1]);
  while ( $word8b ne '' )
    {
    $digitstring = pack ( "a*a*", $digitstring,  sprintf('%03d',ord(unpack ("a", $word8b))));
    $word8b =~ s/^.// ;
    }
  return $digitstring;
  }

sub b_decrypt
  { 
  my $digitstring=$_[1];
  #The strins must have length multiple of 3
  ( $digitstring =~ /^[0-9]*$/ ) || return undef ;
  ( length($digitstring) % 3 == 0 ) || return undef ;
  my $word8b='';
  while ( $digitstring ne '' )
    {
    $word8b = pack ( "a*a*", $word8b,  chr(unpack ("a3", $digitstring)) );
    $digitstring =~ s/^...// ;
    }

  my $pwd=RC4($_[0],$word8b);
  return $pwd;
  }

sub ftp_deltree
  {
  my $ftpcnx=$_[0];
  my $victim=$_[1];

  my $curdir = $ftpcnx->pwd();
  $ftpcnx->cwd($victim);

  my ($to_filelist, $to_file_details)=read_ftp_dir($ftpcnx,$victim);
  my @filelist = @$to_filelist ; 
  my %file_details = %{$to_file_details};

  for my $f (@filelist)
    {
    if ( $file_details{$f}->{rights} =~ /^d/ )
      {
      ftp_deltree($ftpcnx,"$victim/$f"); 
      }
    else
      { 
      print_msg ("    delete $f\n");
      $ftpcnx->delete($f) || print STDERR "Unable to delete file $f.\n"; 
      }
    }

  print_msg ("    rmdir $victim\n");
  $ftpcnx->rmdir($victim);
  my $ls = $ftpcnx->ls($victim) ;
  $ftpcnx->cwd($curdir);
  if ( $ls != '' )
    { 
    print STDERR "Failed to remove dir $victim.\n"; 
    return 0;
    }
  else
    { return 1; }
  }
  
sub read_ftp_dir
  {

  #$string='drwxr-xr-x   2 web      site        16384 Aug  5 16:27 .' ;
  my $RE=   '^(.{10})\s+(\d+)\s+(\w+)\s+(\w+)\s+(\d+)\s+(\w+)\s+(\d+)\s+([^ \t]+)\s+(.+)$';
  

  my $ftpcnx=$_[0];
  my $D=$_[1];
  
  my $curdir=$ftpcnx->pwd();

  $ftpcnx->cwd($D);
  my @RemoteRawFilelist = $ftpcnx->dir;

  my $to_RemoteLongFileList={};
  my $to_RemoteFileList ;  
    {
    my (@RemoteFileList_tmp ) ; 
    foreach my $d ( @RemoteRawFilelist )
      {
      if ( $d =~ /$RE/ )
        { 
        my $F = {} ;
        $F->{rights} = $d ; $F->{rights} =~ s/$RE/$1/ ; 
        $F->{digits} = $d ; $F->{digits} =~ s/$RE/$2/ ; 
        $F->{group} = $d ; $F->{group} =~ s/$RE/$3/ ; 
        $F->{user} = $d ; $F->{user} =~ s/$RE/$4/ ; 
        $F->{size} = $d ; $F->{size} =~ s/$RE/$5/ ; 
        $F->{month} = $d ; $F->{month} =~ s/$RE/$6/ ; 
        $F->{date} = $d ; $F->{date} =~ s/$RE/$7/ ; 
        $F->{time} = $d ; $F->{time} =~ s/$RE/$8/ ; 
        $F->{filename} = $d ; $F->{filename} =~ s/$RE/$9/ ; 
        if ( $F->{filename} ne '.' && $F->{filename} ne '..' )
          {
          if ( defined $to_RemoteLongFileList->{$F->{filename}} ) 
            { print STDERR "Warning: multiple occurence of $to_RemoteLongFileList->{$F->{filename}}\n"; }
          $to_RemoteLongFileList->{$F->{filename}} = $F ;
  #        print "$F->{filename} \n";
        push @RemoteFileList_tmp, $F->{filename} ;
        }
        }
      else
        {
        print STDERR "Warning: undecipherable raw dir line:$d\n";
        }      
      }
    $to_RemoteFileList=[ sort @RemoteFileList_tmp ] ;
    }

  $ftpcnx->cwd($curdir);
  return ($to_RemoteFileList, $to_RemoteLongFileList);
  
  }
